#include <cstdio>
#include <cstring>
#include <algorithm>
#include <string>
#include <windows.h>

using std::string;
using std::exception;

#define eprintf(...) fprintf(stderr, __VA_ARGS__)

// IServer exit codes
const int OK_EXIT_CODE = 0;
const int WA_EXIT_CODE = 1;
const int PE_EXIT_CODE = 2;

class winapi_exception : exception {
  char *message;

  public:
  winapi_exception(const char * const msg) {
    int len = strlen(msg);
    message = new char[len + 30];
    snprintf(message, len + 30, "%s, GetLastError=%x", msg, (unsigned)GetLastError());
  }
  ~winapi_exception() throw() {
    delete[] message;
  }
  virtual const char* what() const throw() {
    return message;
  }
};

void change_inheritance(HANDLE &h, BOOL inherit) {
  HANDLE res;
  if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &res, 0, inherit, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
    throw winapi_exception("Cannot duplicate handle");
  h = res;
}

class subprocess {
  PROCESS_INFORMATION pinfo;
  STARTUPINFO sinfo;

  public:
  static const int KILLED_EXITCODE = 239017;

  // Creates process with command 'cmd' and redirects its standard i/o streams
  // Process is created in suspended state, you have to call 'start()' afterwards
  subprocess(const string &cmd, HANDLE redir_stdin, HANDLE redir_stdout) {
    memset(&sinfo, 0, sizeof sinfo);
    sinfo.cb = sizeof(sinfo);
    sinfo.dwFlags |= STARTF_USESTDHANDLES; // Enable standard handles redirection

    sinfo.hStdInput = redir_stdin;
    sinfo.hStdOutput = redir_stdout;
    sinfo.hStdError = GetStdHandle(STD_ERROR_HANDLE);

    // This error mode is inherited by created process
    // Avoids standard windows error messages which results to process hang
    DWORD last_error_mode = SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX); 
    if (!CreateProcess(0, (LPSTR)cmd.c_str(), NULL, NULL, TRUE, CREATE_SUSPENDED, NULL, NULL, &sinfo, &pinfo))
      throw winapi_exception("Cannot create process");
    SetErrorMode(last_error_mode); // Restore old mode for this app
  }
  void start() {
    if (!ResumeThread(pinfo.hThread))
      throw winapi_exception("Cannot resume process' thread");
  }
  HANDLE get_process_handle() const { return pinfo.hProcess; }

  // This function can return STILL_ACTIVE (259)
  DWORD get_exit_code() const {
    DWORD res;
    if (!GetExitCodeProcess(get_process_handle(), &res))
      throw winapi_exception("Cannot GetExitCodeProcess");
    return res;
  }
  ~subprocess() {
    TerminateProcess(pinfo.hProcess, KILLED_EXITCODE);
  }
};

// Waits for process and returns true
// or return false if it doesn't consume CPU time
// during 100 ms (and, probably, locked by input operations)
bool wait_for_process_or_il(HANDLE hProcess) {
  const DWORD poll_interval = 20; // we query process' time counters every 50ms
  const int MAX_IDLE_TICKS = 2; // if these counters are stable for 50*2=100ms, we consider this as IL

  FILETIME creation_time, exit_time; // don't really needed

  FILETIME last_kernel_time, last_user_time;
  memset(&last_kernel_time, 0, sizeof(last_kernel_time));
  memset(&last_user_time, 0, sizeof(last_user_time));

  FILETIME kernel_time, user_time;
  int idle_ticks = 0;

  for (;;) {
    DWORD res = WaitForSingleObject(hProcess, poll_interval); // Wait for process termination
    if (res == WAIT_OBJECT_0) return true;
    if (res != WAIT_TIMEOUT)
      throw winapi_exception("Cannot WaitForSingleObject (waiting for IL)");

    if (!GetProcessTimes(hProcess, &creation_time, &exit_time, &kernel_time, &user_time))
      throw winapi_exception("Cannot GetProcessTimes");

    // Compare old and new counters
    if (memcmp(&last_kernel_time, &kernel_time, sizeof kernel_time) ||
        memcmp(&last_user_time, &user_time, sizeof user_time)) {
      last_kernel_time = kernel_time;
      last_user_time = user_time;
      idle_ticks = 0;
      continue;
    }
    if (++idle_ticks >= MAX_IDLE_TICKS)
      return false;
  } 
}

int main(int argc, char* argv[]) {
  if (argc < 3) {
    eprintf("Usage: runi.exe <solution command> <interactor command>\n");
    return 1;
  }

  const string solution_cmd = argv[1];
  const string interactor_cmd = argv[2];

  HANDLE i2s_read, i2s_write;
  HANDLE s2i_read, s2i_write;
  if (!CreatePipe(&i2s_read, &i2s_write, NULL, 0))
    throw winapi_exception("Cannot create pipe");
  if (!CreatePipe(&s2i_read, &s2i_write, NULL, 0))
    throw winapi_exception("Cannot create pipe");

  eprintf("Creating interactor...\n");
  change_inheritance(s2i_read, TRUE);
  change_inheritance(i2s_write, TRUE);
  subprocess interactor(interactor_cmd, s2i_read, i2s_write);
  // Avoid inheritance of these handles by the second process
  change_inheritance(s2i_read, FALSE);
  change_inheritance(i2s_write, FALSE);


  eprintf("Creating solution...\n");
  change_inheritance(i2s_read, TRUE);
  change_inheritance(s2i_write, TRUE);
  subprocess solution(solution_cmd, i2s_read, s2i_write);
  change_inheritance(i2s_read, FALSE);
  change_inheritance(s2i_write, FALSE);

  eprintf("Starting interactor...\n");
  interactor.start();
  eprintf("Starting solution...\n");
  solution.start();

  HANDLE objects[] = { interactor.get_process_handle(), solution.get_process_handle() };
  DWORD wait_res = WaitForMultipleObjects(2, objects, FALSE, INFINITE);

  switch (wait_res) {
  case WAIT_OBJECT_0: // interactor
    eprintf("Interactor terminated\n");
    CloseHandle(i2s_write);
    CloseHandle(i2s_read);

    switch (interactor.get_exit_code()) {
    case WA_EXIT_CODE:
      printf("WA\n");
      break;

    case PE_EXIT_CODE:
      printf("PE\n");
      break;

    default:
      printf("JE\n");
      break;
    }
    break;

  case WAIT_OBJECT_0 + 1: // solution
    eprintf("Solution terminated\n");
    if (solution.get_exit_code() == 0) {
      CloseHandle(s2i_write);
      CloseHandle(s2i_read);

       wait_res = WaitForSingleObject(interactor.get_process_handle(), INFINITE);
       if (wait_res != WAIT_OBJECT_0)
         throw winapi_exception("Error occured during WaitForSingleObject (interactor)\n");
       switch (interactor.get_exit_code()) {
       case OK_EXIT_CODE:
         printf("OK\n");
         break;

       case WA_EXIT_CODE:
         printf("WA\n");
         break;

       case PE_EXIT_CODE:
         printf("PE\n");
         break;

       default:
         printf("JE\n");
         break;
       }
     } else { // That's probably RTE/WA
       if (!wait_for_process_or_il(interactor.get_process_handle())) {
         printf("RTE\n");
         break;
       } else {
         switch (interactor.get_exit_code()) {
         case OK_EXIT_CODE:
           printf("RTE\n");
           break;

         case WA_EXIT_CODE:
           printf("WA\n");
           break;

         case PE_EXIT_CODE:
           printf("PE\n");
           break;

         default:
           printf("JE\n");
           break;
         }
       }
     }
    break;  

  default:
    throw winapi_exception("Error occured during WaitForMultipleObjects");
  }

  return 0;
}
